"""
Tests for distribution module
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey


import unittest
import numpy as np
from numpy.testing import assert_almost_equal, assert_equal
from uravu.distribution import Distribution
from scipy.stats import norm, uniform

class TestDistribution(unittest.TestCase):
    """
    Testing the Distribution class.
    """
    def test_init_name(self):
        distro = Distribution(norm.rvs(loc=0, scale=1, size=1000), name='random')
        assert_equal(distro.name, 'random')

    def test_init_samples(self):
        samples = norm.rvs(loc=0, scale=1, size=1000)
        distro = Distribution(samples)
        assert_equal(distro.samples, samples)

    def test_init_ci_points(self):
        distro = Distribution(norm.rvs(loc=0, scale=1, size=1000))
        assert_equal(distro.ci_points, [2.5, 97.5]) 
    
    def test_init_ci_points_optional(self):
        distro = Distribution(norm.rvs(loc=0, scale=1, size=1000), ci_points=[5, 95])
        assert_equal(distro.ci_points, [5, 95]) 

    def test_init_ci_points_error(self):
        with self.assertRaises(ValueError):
            distro = Distribution(norm.rvs(loc=0, scale=1, size=1000), ci_points=[5, 50, 95])

    def test_normal(self):
        distro = Distribution(norm.rvs(loc=0, scale=1, size=100, random_state=np.random.RandomState(1)))
        assert_equal(distro.normal, True)

    def test_not_normal(self):
        distro = Distribution(uniform.rvs(loc=0, scale=1, size=100, random_state=np.random.RandomState(1)))
        assert_equal(distro.normal, False)

    def test_size(self):
        distro = Distribution(uniform.rvs(loc=0, scale=1, size=100))
        assert_equal(distro.size, 100)

    def test_pdf(self):
        distro = Distribution(uniform.rvs(loc=0, scale=1, size=100, random_state=np.random.RandomState(1)))
        assert_almost_equal(distro.pdf(0), [0.5805], decimal=4)

    def test_logpdf(self):
        distro = Distribution(uniform.rvs(loc=0, scale=1, size=100, random_state=np.random.RandomState(1)))
        assert_almost_equal(distro.logpdf(0), np.log([0.5805]), decimal=4)

    def test_negative_pdf(self):
        distro = Distribution(uniform.rvs(loc=0, scale=1, size=100, random_state=np.random.RandomState(1)))
        assert_almost_equal(distro.negative_pdf(0), [-0.5805], decimal=4)

    def test_dist_max(self):
        distro = Distribution(norm.rvs(loc=0, scale=1, size=100, random_state=np.random.RandomState(1)))
        assert_almost_equal(distro.dist_max, 0, decimal=1) 

    def test_min(self):
        distro = Distribution(np.linspace(1, 10, 100))
        assert_equal(distro.min, 1)
    
    def test_max(self):
        distro = Distribution(np.linspace(1, 10, 100))
        assert_equal(distro.max, 10)

    def test_n(self):
        distro = Distribution(norm.rvs(loc=0, scale=1, size=100, random_state=np.random.RandomState(1)))
        assert_almost_equal(distro.n, 0, decimal=1)

    def test_s(self):
        distro = Distribution(norm.rvs(loc=0, scale=1, size=100, random_state=np.random.RandomState(2)))
        assert_almost_equal(distro.s, 1, decimal=1)

    def test_v(self):
        distro = Distribution(norm.rvs(loc=0, scale=2, size=100, random_state=np.random.RandomState(2)))
        assert_almost_equal(distro.v, 4, decimal=0)

    def test_s_uniform(self):
        distro = Distribution(uniform.rvs(loc=0, scale=1, size=100, random_state=np.random.RandomState(2)))
        assert_equal(distro.s, None)

    def test_v_uniform(self):
        distro = Distribution(uniform.rvs(loc=0, scale=2, size=100, random_state=np.random.RandomState(2)))
        assert_equal(distro.v, None)

    def test_con_int(self):
        distro = Distribution(norm.rvs(loc=0, scale=2, size=10000, random_state=np.random.RandomState(2)))
        assert_almost_equal(distro.con_int[0], -2 * 1.96, decimal=1)
        assert_almost_equal(distro.con_int[1], 2 * 1.96, decimal=1)

    def test_add_samples(self):
        samples = norm.rvs(loc=0, scale=2, size=100, random_state=np.random.RandomState(2))
        distro = Distribution(samples)
        distro.add_samples(samples)
        assert_equal(distro.size, 200)


"""
These are plotting functions that take either :py:class:`~uravu.relationship.Relationship` or :py:class:`~uravu.distribution.Distribution` class objects.

The aim is to produce *publication quality* plots. However, we recognise that taste exists, and ours may be different from yours. The colorscheme in this work was chosen to be colorblind friendly.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

import numpy as np

try:
    import matplotlib.pyplot as plt
    from corner import corner
    from uravu import _fig_params
    colors = _fig_params.colors
except ModuleNotFoundError:
    raise ModuleNotFoundError("The matplotlib, corner and seaborn packages are necessary for the use of the plotting module, please install these.")


def plot_relationship(relationship, axes=None, figsize=(10, 6)):  # pragma: no cover
    """
    Plot the relationship. Additional plots will be included on this if posterior sampling has been used to find distributions.

    Args:
        relationship (:py:class:`uravu.relationship.Relationship`): The relationship to be plotted.
        axes (:py:class:`matplotlib.axes.Axes`, optional): Axes to which the plot should be added. If :py:attr:`None` given new axes will be created. Default is :py:attr:`None`.
        fig_size (:py:attr:`tuple`, optional): horizontal and veritcal size for figure (in inches). Default is :py:attr:`(10, 6)`.

    Returns:
        (:py:class:`matplotlib.axes.Axes`): The axes with new plots.
    """
    if axes is None:
        axes = plt.subplots(figsize=figsize)[1]
    variables = relationship.variables
    axes.errorbar(relationship.x, relationship.y.mode, relationship.y.s, c=colors[0], ecolor=colors[0] + '40', marker='.', ls='')
    smooth_x = np.linspace(relationship.x.min(), relationship.x.max(), 1000)
    if relationship.mcmc_done or relationship.nested_sampling_done:
        plot_samples = np.random.randint(0, variables[0].samples.size, size=100)
        for i in plot_samples:
            float_variables = relationship.get_sample(i)
            axes.plot(smooth_x, relationship.function(smooth_x, *float_variables), color=colors[1], alpha=0.05)
    else:
        float_variables = relationship.variable_medians
        axes.plot(smooth_x, relationship.function(smooth_x, *float_variables), color=colors[1])
    return axes


def plot_distribution(distro, axes=None, figsize=(5, 3)):  # pragma: no cover
    """
    Plot the probability density function for a distribution.

    Args:
        distro (:py:class`uravu.distriobution.Distribution`): The distribution to be plotted.
        axes (:py:class:`matplotlib.axes.Axes`, optional): Axes to which the plot should be added. If :py:attr:`None` given new axes will be created. Default is :py:attr:`None`.
        fig_size (:py:class:`tuple`): Horizontal and veritcal size for figure (in inches). Default is :py:attr:`(10, 6)`.

    Returns:
        (:py:class:`matplotlib.axes.Axes`): The axes with new plots.
    """
    if axes is None:
        axes = plt.subplots(figsize=figsize)[1]
    kde = distro.kde
    abscissa = np.linspace(distro.samples.min(), distro.samples.max(), 100)
    ordinate = kde.evaluate(abscissa)
    axes.plot(abscissa, ordinate, color=colors[0])
    axes.hist(distro.samples, bins=25, density=True, color=colors[0], alpha=0.5)
    axes.fill_betweenx(np.linspace(0, ordinate.max() + ordinate.max() * 0.1), distro.con_int[0], distro.con_int[1], alpha=0.2)
    axes.set_ylim((0, ordinate.max() + ordinate.max() * 0.1))
    return axes


def plot_corner(relationship, figsize=(8, 8)):  # pragma: no cover
    """
    Plot the :py:mod:`corner` (named for the Python package) plot between the relationships variables.

    Args:
        relationship (:py:class:`uravu.relationship.Relationship`): The relationship containing the distributions to be plotted.
        fig_size (:py:attr:`tuple`, optional): horizontal and veritcal size for figure (in inches). Default is :py:attr:`(10, 6)`.

    Returns:
        :py:attr:`tuple`: Containing:
            - :py:class:`matplotlib.figure.Figure`: The figure with new plots.
            - :py:class:`matplotlib.axes.Axes`: The axes with new plots.
    """
    n = len(relationship.variables)
    fig, ax = plt.subplots(n, n, figsize=figsize)
    samples = np.zeros((relationship.variables[0].size, len(relationship.variables)))
    for i, v in enumerate(relationship.variables):
        samples[:, i] = v.samples
    corner(samples, color=colors[0], hist_kwargs={"lw": 4, "histtype": "step"}, label_kwargs={"fontsize": _fig_params.rcParams["axes.labelsize"]}, fig=fig)
    return fig, ax



"""
A few additional utility functions to improve the usability of :py:mod:`uravu`.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey


import numpy as np
from scipy.stats import pearsonr


def straight_line(abscissa, gradient, intercept):
    """
    A one dimensional straight line function.

    Args:
        abscissa (:py:attr:`array_like`): The abscissa data.
        gradient (:py:attr:`float`): The slope of the line.
        intercept (:py:attr:`float`): The y-intercept of the line.

    Returns:
        :py:attr:`array_like`: The resulting ordinate.
    """
    return gradient * abscissa + intercept


def bayes_factor(model_1, model_2):
    """
    Use the `Bayes factor`_ to compare two models. Using Table from `Kass and Raftery`_ to compare.

    Args:
        model_1 (:py:class:`uncertainties.core.Variable` or :py:attr:`float`): ln evidence for model 1.
        model_2 (:py:class:`uncertainties.core.Variable` or :py:attr:`float`): ln evidence for model 2.

    Return:
        :py:class:`uncertainties.core.Variable` or :py:attr:`float`: 2ln(B), where B is the Bayes Factor between the two models.

    .. _Bayes factor: https://en.wikipedia.org/wiki/Bayes_factor
    .. _Kass and Raftery: https://www.colorado.edu/amath/sites/default/files/attached-files/kassraftery95.pdf
    """
    return 2 * (model_1 - model_2)


def correlation_matrix(relationship):
    """
    Evaluate the `Pearsons correlation coefficient`_ matrix for the variables in a given relationship.

    Args:
        relationship (:py:class:`uravu.relationship.Relationship`): The relationship to determine the correlation matrix for.

    Returns:
        :py:attr:`array_like`: The correlation matrix for the relationships variables.

    .. _Pearsons correlation coefficient: https://en.wikipedia.org/wiki/Pearson_correlation_coefficient
    """
    n = len(relationship.variables)
    matrix = np.zeros((n, n))
    subset = np.random.randint(0, relationship.variables[0].size, size=400)
    for i in range(n):
        samples_i = relationship.variables[i].samples[subset]
        for j in range(i, n):
            samples_j = relationship.variables[j].samples[subset]
            matrix[i, j] = pearsonr(samples_i, samples_j)[0]
            matrix[j, i] = pearsonr(samples_j, samples_i)[0]
    return matrix



"""
The optimize module includes the functionality necessary for maximum likelihood determination.
Furthermore, the natural log likelihood function used in the :func:`~uravu.sampling.mcmc()` and :func:`~uravu.sampling.nested_sampling()` methods may be found here.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

import numpy as np
from scipy.optimize import minimize, differential_evolution


def max_ln_likelihood(relationship, method, x0=None, **kwargs):
    """
    Determine the variable values which maximize the likelihood for the given relationship. For keyword arguments see the :func:`scipy.optimize.minimize()` documentation.

    Args:
        relationship (:py:class:`uravu.relationship.Relationship`): The relationship for which variables should be found.
        x0 (:py:attr:`array_like`, optional): Initial guesses for the variable values. Default to the current :py:attr:`~uravu.relationship.Relationship.variables` values which are initialized as all :py:attr:`1`.

    Return:
        :py:attr:`array_like`: Optimized variables for the relationship.
    """
    if x0 is None:
        x0 = relationship.variable_medians
    args = (relationship.function, relationship.abscissa, relationship.ordinate)
    if method == 'diff_evo':
        res = differential_evolution(negative_lnl, relationship.bounds, args=args, **kwargs)
    elif method == 'mini':
        res = minimize(negative_lnl, x0, args=args, bounds=relationship.bounds, **kwargs)
    return res.x


def negative_lnl(variables, function, abscissa, ordinate):
    """
    Calculate the negative natural logarithm of the likelihood given a set of variables, when there is no uncertainty in the abscissa.

    Args:
        variables (:py:attr:`array_like`): Variables for the function evaluation.
        function (:py:attr:`callable`): The function to be evaluated.
        abscissa (:py:attr:`array_like`): The abscissa values.
        ordinate (:py:attr:`array_like`): The ordinate values.

    Returns:
        :py:attr:`float`: Negative natural log-likelihood between model and data.
    """
    return -ln_likelihood(variables, function, abscissa, ordinate)


def ln_likelihood(variables, function, abscissa, ordinate):
    """
    Calculate the natural logarithm of the likelihood given a set of variables, when there is no uncertainty in the abscissa.

    Args:
        variables (:py:attr:`array_like`): Variables for the function evaluation.
        function (:py:attr:`callable`): The function to be evaluated.
        abscissa (:py:attr:`array_like`): The abscissa values.
        ordinate (:py:attr:`array_like`): The ordinate values.

    Returns:
         :py:attr:`float`: Natural log-likelihood between model and data.
    """
    model = function(abscissa, *variables)
    ln_l = ordinate.logpdf(model)
    return np.sum(ln_l)



"""
The :py:class:`~uravu.relationship.Relationship` class is a foundational component of the :py:mod:`uravu` package, and acts as the main API for use of the package.
This class enables the storage of the relationship between the model and the data.

Objects of this class offer easy methods to perform maximum likelihood evaluation, Markiv chain Monte Carlo (MCMC) for posterior probabiltiy determination and Bayesian evidence estimation by nested sampling.

See the `tutorials online`_ for more guidence of how to use this package.

.. _tutorials online: https://uravu.rtfd.io
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

from inspect import getfullargspec
import numpy as np
from scipy import stats
from uncertainties import ufloat
from uravu import optimize, sampling
from uravu.distribution import Distribution
from uravu.axis import Axis


class Relationship:
    """
    The :py:class:`~uravu.relationship.Relationship` class is the base of the :py:mod:`uravu` package, enabling the use of Bayesian inference for the assessment of a model's ability to describe some data.

    Attributes:
        function (:py:attr:`callable`): The function that is modelled.
        abscissa (:py:attr:`array_like`): The abscissa data that the modelling should be performed on.
        ordinate (:py:attr:`list` or :py:class:`uravu.distribution.Distribution` or :py:attr:`array_like`): The ordinate data against with the model should be compared. This should be an :py:attr:`list` or :py:class:`uravu.distribution.Distribution` unless a :py:attr:`ordinate_error` is given.
        variables (:py:attr:`list` of :py:class:`uravu.distribution.Distribution`): Variables in the :py:attr:`~uravu.relationship.Relationship.function`.
        bounds (:py:attr:`tuple`): The minimum and maximum values for each parameters.
        ln_evidence (:py:class:`uncertainties.core.Variable`): The natural-log of the Bayesian evidence for the model to the given data.
        mcmc_results (:py:attr:`dict`): The results from :func:`emcee.EnsembleSampler.run_mcmc()` sampling.
        nested_sampling_results (:py:attr:`dict`): The results from :func:`dynesty.NestedSampler.run_nested()` nested sampling.

    Args:
        function (:py:attr:`callable`): The functional relationship to be modelled.
        abscissa (:py:attr:`array_like`): The abscissa data. If multi-dimensional, the array is expected to have the shape :py:attr:`(N, d)`, where :py:attr:`N` is the number of data points and :py:attr:`d` is the dimensionality.
        ordinate (:py:attr:`list` or :py:class:`uravu.distribution.Distribution` or :py:attr:`array_like`): The ordinate data. This should have a shape :py:attr:`(N,)`.
        bounds (:py:attr:`tuple`, optional): The minimum and maximum values for each parameters. Defaults to :py:attr:`None`.
        ordinate_error (:py:attr:`array_like`, optional): The uncertainty in the ordinate. Only used if :py:attr:`ordinate` is not a :py:attr:`list` or :py:class:`uravu.distribution.Distribution`. Defaults to :py:attr:`None`.
    """

    def __init__(self, function, abscissa, ordinate, bounds=None, ordinate_error=None):
        """
        Initialisation function for a :py:class:`~uravu.relationship.Relationship` object.
        """
        self.function = function
        self.abscissa = abscissa

        potential_y = []
        for i, y in enumerate(ordinate):
            if not isinstance(y, Distribution):
                if not isinstance(y, stats._distn_infrastructure.rv_frozen):
                    if ordinate_error is not None:
                        potential_y.append(Distribution(stats.norm.rvs(loc=y, scale=ordinate_error[i], size=5000)))
                    else:
                        raise ValueError("uravu ordinate should be a list of uravu.distribution.Distribution objects or an ordinate_error should be given.")
                else:
                    potential_y.append(Distribution(y.rvs(size=5000)))
                self.ordinate = Axis(potential_y)
            else:
                self.ordinate = Axis(ordinate)

        if abscissa.shape[0] != len(ordinate):
            raise ValueError("The number of data points in the abscissa does not match that for the ordinate.")

        self.bounds = bounds
        self.variables = []
        if bounds is not None:
            if len(self.bounds) != self.len_parameters or not isinstance(bounds[0], tuple):
                raise ValueError("The number of bounds does not match the number of parameters")
            for i, b in enumerate(self.bounds):
                self.variables.append(Distribution(stats.uniform.rvs(loc=b[0], scale=b[1] - b[0], size=500)))
        else:
            for i in range(self.len_parameters):
                self.variables.append(Distribution(1))

        self.ln_evidence = None
        self.mcmc_results = None
        self.nested_sampling_results = None

    @property
    def x(self):
        """
        Abscissa values.

        Returns:
            :py:attr:`array_like`: Abscissa values.
        """
        return self.abscissa

    @property
    def y(self):
        """
        Ordinate values.

        Returns:
            :py:attr:`array_like`: Ordinate values.
        """
        return self.ordinate

    @property
    def variable_medians(self):
        """
        The median values for each of the variables.

        Returns:
            :py:attr:`array_like`: Variable medians.
        """
        medians = np.zeros((len(self.variables)))
        for i, var in enumerate(self.variables):
            medians[i] = var.n
        return medians

    @property
    def variable_modes(self):
        """
        The mode values for each of the variables.

        Returns:
            :py:attr:`array_like`: Variable modes.
        """
        modes = np.zeros((len(self.variables)))
        for i, var in enumerate(self.variables):
            modes[i] = var.dist_max
        return modes

    @property
    def mcmc_done(self):
        """
        Has MCMC been performed? Determined based on the type of the variables.

        Returns:
            :py:attr:`bool`: Has MCMC been performed?
        """
        if self.mcmc_results is not None:
            return True
        return False

    @property
    def nested_sampling_done(self):
        """
        Has nested sampling been performed? Determined based on if the ln_evidence has a value.

        Returns:
            :py:attr:`bool`: Has nested sampling been performed?
        """
        if self.ln_evidence is not None:
            return True
        return False

    def get_sample(self, i):
        """
        Return the variable values for a given sample.

        Args:
            i (:py:attr:`int`): The sample index.

        Returns:
            :py:attr:`list` of :py:attr:`float`: Variable values at given index.
        """
        return [self.variables[j].samples[i] for j in range(self.len_parameters)]

    @property
    def len_parameters(self):
        """
        Determine the number of variables in the assessment function.

        Returns:
            :py:attr`int`: Number of variables.
        """
        # The minus one is to remove the abscissa data which is a
        # argument in the assessment function
        return len(getfullargspec(self.function).args) - 1

    def bayesian_information_criteria(self):
        """
        Calculate the `Bayesian information criteria`_ for the relationship.

        Returns:
            :py:attr:`float`: Bayesian information criteria.

        .. _Bayesian information criteria: https://en.wikipedia.org/wiki/Bayesian_information_criterion
        """
        self.max_likelihood('diff_evo')
        l_hat = optimize.ln_likelihood(self.variable_medians, self.function, self.abscissa, self.ordinate)
        return np.log(self.x.size) * self.len_parameters - 2 * l_hat

    def max_likelihood(self, method, x0=None, **kwargs):
        """
        Determine values for the variables which maximise the likelihood for the :py:class:`~uravu.relationship.Relationship`. For keyword arguments see the :func:`scipy.optimize.minimize()` documentation.

        Args:
            x0 (:py:attr:`array_like`): Initial guess values for the parameters.
        """
        var = optimize.max_ln_likelihood(self, method, x0, **kwargs)
        for i, v in enumerate(var):
            self.variables[i] = Distribution(v)

    def prior(self):
        """
        *Standard priors* for the relationship. These priors are broad, uninformative, for normal variables running the range :py:attr:`[x - x * 10, x + x * 10)` (where :py:attr:`x` is the variable value).

        Returns:
            :py:attr:`list` of :py:class:`scipy.stats.rv_continuous`: :py:mod:`scipy.stats` functions describing the priors.
        """
        priors = []
        if self.bounds is not None:
            for i, var in enumerate(self.variable_medians):
                loc = self.bounds[i][0]
                scale = self.bounds[i][1] - loc
                priors.append(stats.uniform(loc=loc, scale=scale))
        else:
            for var in self.variable_medians:
                loc = var - 10
                scale = (var + 10) - loc
                priors.append(stats.uniform(loc=loc, scale=scale))
        return priors

    def mcmc(self, prior_function=None, walkers=50, n_samples=500, n_burn=500, progress=True):
        """
        Perform MCMC to get the posterior probability distributions for the variables of the relationship. *Note*, running this method will populate the :py:attr:`~uravu.relationship.Relationship.variables` attribute with :py:class:`~uravu.distribution.Distribution` objects. Once run, a result dictionary containing the :py:attr:`distributions`, :py:attr:`chain`, and :py:attr:`samples` from :py:mod:`emcee` is piped into the class variable :py:attr:`mcmc_results`. 

        Args:
            prior_function (:py:attr:`callable`, optional): The function to populated some prior distributions. Default is the broad uniform priors in :func:`~uravu.relationship.Relationship.prior()`.
            walkers (:py:attr:`int`, optional): Number of MCMC walkers. Default is :py:attr:`50`.
            n_samples (:py:attr:`int`, optional): Number of sample points. Default is :py:attr:500`.
            n_burn (:py:attr:`int`, optional): Number of burn in samples. Default is :py:attr:`500`.
            progress (:py:attr:`bool`, optional): Show :py:mod:`tqdm` progress for sampling. Default is :py:attr:`True`.
        """
        self.mcmc_results = sampling.mcmc(self, prior_function=prior_function, walkers=walkers, n_samples=n_samples, n_burn=n_burn, progress=progress)
        self.variables = self.mcmc_results["distributions"]

    def nested_sampling(self, prior_function=None, progress=True, dynamic=False, **kwargs):
        """
        Perform nested sampling, or dynamic nested sampling, to determine the Bayesian natural-log evidence. For keyword arguments see the :func:`dynesty.NestedSampler.run_nested()` documentation. Once run, the result dictionary produced by :func:`dynesty.NestedSampler.run_nested()` is piped into the class variable :py:attr:`nested_sampling_results`. 

        Args:
            prior_function (:py:attr:`callable`, optional): The function to populate some prior distributions. Default is the broad uniform priors in :func:`~uravu.relationship.Relationship.prior()`.
            progress (:py:attr:`bool`, optional): Show :py:mod:`tqdm` progress for sampling. Default is :py:attr:`True`.
        """
        self.nested_sampling_results = sampling.nested_sampling(self, prior_function=prior_function, progress=progress, dynamic=dynamic, **kwargs)
        self.ln_evidence = ufloat(self.nested_sampling_results["logz"][-1], self.nested_sampling_results["logzerr"][-1])
        self.variables = self.nested_sampling_results["distributions"]



"""
The storage and manipulation of probability distributions is fundamental to the operation of ``uravu`` and Bayesian inference.
The :py:class:`~uravu.distribution.Distribution` class oversees these operations.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

import numpy as np
from scipy.stats import normaltest, gaussian_kde
from scipy.optimize import minimize


class Distribution:
    """
    In addition to storage of the probability distribution, this class allows for some basic analysis, such as determination of normality.

    Attributes:
        samples (:py:attr:`array_like`): Samples in the distribution.
        name (:py:attr:`str`): Distribution name.
        ci_points (:py:attr:`array_like`): The percentiles at which confidence intervals should be found.
        normal (:py:attr:`bool`): Are the samples normally distributed?
        kde (:py:class:`scipy.stats.kde.gaussian_kde`): Kernel density approximation for the distribution.

    Args:
        samples (:py:attr:`array_like`): Sample for the distribution.
        name (:py:attr:`str`, optional): A name to identify the distribution. Default is :py:attr:`'Distribution'`.
        ci_points (:py:attr:`array_like`, optional): The two percentiles at which confidence intervals should be found. Default is :py:attr:`[2.5, 97.5]` (a 95 % confidence interval).

    .. _FAQ: ./faq.html
    """

    def __init__(self, samples, name="Distribution", ci_points=None):
        """
        Initialisation function for a :py:class:`~uravu.distribution.Distribution` object.
        """
        self.name = name
        self.samples = np.array([])
        if ci_points is None:
            self.ci_points = np.array([2.5, 97.5])
        else:
            if len(ci_points) != 2:
                raise ValueError("The ci_points must be an array of length two.")
            self.ci_points = np.array(ci_points)
        self.normal = False
        self.add_samples(np.array(samples))

    @property
    def size(self):
        """
        Get the number of samples in the distribution.

        Returns:
            :py:attr:`int`: Number of samples.
        """
        return self.samples.size

    def check_normality(self):
        """
        Uses a :func:`scipy.stats.normaltest()` to evaluate if samples are normally distributed and updates the :py:attr:`~uravu.distribution.Distribution.normal` attribute.
        """
        alpha = 0.05
        test_samples = self.samples
        if self.size > 500:
            test_samples = np.random.choice(self.samples, size=500)
        p_value = normaltest(test_samples)[1]
        if p_value > alpha:
            self.normal = True
        else:
            self.normal = False

    def pdf(self, x):
        """
        Get the probability density function for the distribution.

        Args:
            x (:py:attr:`float`): Value to return probability of.

        Return:
            :py:attr:`float`: Probability.
        """
        return self.kde.pdf(x)

    def logpdf(self, x):
        """
        Get the natural log probability density function for the distribution.

        Args:
            x (:py:attr:`float`): Value to return natural log probability of.

        Return:
            :py:attr:`float`: Natural log probability.
        """
        return self.kde.logpdf(x)

    def negative_pdf(self, x):
        """
        Get the negative of the probability density function for the distribution.

        Args:
            x (:py:attr:`float`): Value to return negative probability of.

        Return:
            :py:attr:`float`: Negative probability.
        """
        return -self.kde.pdf(x) 
        
    @property
    def dist_max(self):
        """
        Get the value that maximises the distribution. If no :py:attr:`kde` has been created (for example if the distribution has fewer than 8 values) the median is returned. 
        
        Returns
            :py:attr:`float`: Most likely value.
        """
        try: 
            return minimize(self.negative_pdf, x0=[self.n]).x
        except AttributeError:
            return self.n


    @property
    def min(self):
        """
        Get sample minimum.

        Returns:
            :py:attr:`float`: Sample minimum.
        """
        return self.samples.min()

    @property
    def max(self):
        """
        Get sample maximum.

        Returns:
            :py:attr:`float`: Sample maximum.
        """
        return self.samples.max()

    @property
    def n(self):
        """
        Get the median value of the distribution (for a normal distribution this is the same as the mean).

        Returns:
            :py:attr:`float`: Median value.
        """
        return np.percentile(self.samples, [50])[0]

    @property
    def s(self):
        """
        Get the standard deviation of the distribution. For a non-normal distribution, this will return :py:attr:`None`.

        Returns:
            :py:attr:`float` or :py:attr:`None`: Standard deviation of the distribution.
        """
        if self.normal:
            return np.std(self.samples, ddof=1)
        return None

    @property
    def v(self):
        """
        Get the variance of the distribution. For a non-normal distribution, this will return :py:attr:`None`.

        Returns:
            :py:attr:`float` or :py:attr:`None`: Standard deviation of the distribution.
        """
        if self.normal:
            return np.var(self.samples, ddof=1)
        return None

    @property
    def con_int(self):
        """
        Get the extrema of the confidence intervals of the distribution.

        Returns:
            :py:attr:`array_like`: Distribution values at the confidence interval.
        """
        return np.percentile(self.samples, self.ci_points)

    def add_samples(self, samples):
        """
        Add samples to the distribution.

        Args:
            samples (:py:attr:`array_like`): Samples to be added to the distribution.
        """
        self.samples = np.append(self.samples, np.array(samples).flatten())
        if self.size > 8:
            self.check_normality()
            self.kde = gaussian_kde(self.samples)



"""
Tests for optimize module
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

import unittest
import numpy as np
from numpy.testing import assert_almost_equal
import scipy.stats
from uravu import optimize, utils, relationship
from uravu.distribution import Distribution

TEST_Y = []
for i in np.arange(1, 9, 1):
    TEST_Y.append(Distribution(scipy.stats.norm.rvs(loc=i, scale=0.5, size=200)))
TEST_X = np.arange(1, 9, 1)

class TestOptimize(unittest.TestCase):
    """
    Unit tests for optimize module.
    """

    def test_ln_likelihood(self):
        test_rel = relationship.Relationship(
            utils.straight_line, TEST_X, TEST_Y 
        )
        expected_lnl = -3
        actual_lnl = optimize.ln_likelihood(
            [1., 0.], test_rel.function, test_rel.x, test_rel.y
        )
        assert_almost_equal(actual_lnl, expected_lnl, decimal=0)

    def test_negative_lnl(self):
        test_rel = relationship.Relationship(
            utils.straight_line, TEST_X, TEST_Y 
        )
        expected_negtive_lnl = 3
        actual_negative_lnl = optimize.negative_lnl(
            [1., 0.], test_rel.function, test_rel.x, test_rel.y
        )
        assert_almost_equal(actual_negative_lnl, expected_negtive_lnl, decimal=0)

    def test_max_lnlikelihood(self):
        test_rel = relationship.Relationship(
            utils.straight_line, TEST_X, TEST_Y 
        )
        actual_best_variables = optimize.max_ln_likelihood(test_rel, 'mini')
        assert_almost_equal(actual_best_variables, np.array([1, 0]), decimal=0)



"""
The :py:class:`~uravu.axis.Axis` class controls the organisation of axes in the uravu code, including the evaluation of an axis-level multidimensional kernal density estimate.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey


import numpy as np
from scipy.stats import gaussian_kde
from uravu.distribution import Distribution


class Axis:
    """
    The Axes class is a flexible storage option for both numerical (:py:class:`numpy.ndarray`) and distribution (:py:class:`uravu.distribution.Distribution`) arrays.

    Attributes:
        values (:py:attr:`list` or :py:class:`uravu.distribution.Distribution` or :py:attr:`array_like`): Array of values.
        kde (:py:class:`scipy.stats.kde.gaussian_kde`): Multi-dimensional kernel density approximation for the axes.

    Args:
        values (:py:attr:`list` or :py:class:`uravu.distribution.Distribution` or :py:attr:`array_like`): Array of values.
    """
    def __init__(self, values):
        """
        Initialisation function for a :py:class:`~uravu.axis.Axis` object.
        """
        self.kde = None
        if isinstance(values[0], Distribution):
            self.values = values
            self.kde = _get_kde(self.values)
        else:
            self.values = np.array(values)
        
    @property
    def n(self):
        """
        Get the median of each value in the axis.

        Returns:
            :py:attr:`array_like`: Medians for axis.
        """
        v = np.zeros(self.shape)
        if isinstance(self.values[0], Distribution):
            for i, o in enumerate(self.values):
                v[i] = o.n
            return v
        return self.values

    @property
    def s(self):
        """
        Get the uncertainty from confidence intervals of each value in the axis.

        Returns:
            :py:attr:`array_like`: Uncertainties for axis.
        """
        if isinstance(self.values[0], Distribution):
            dv = np.zeros((2, self.size))
            for i, o in enumerate(self.values):
                dv[:, i] = np.abs(o.con_int - o.n)
            return dv
        return np.zeros(self.shape)

    @property
    def mode(self):
        """
        Get the values that maximise the probability distributions of the axis. 

        Returns:
            :py:attr:`array_like`: Values that maximise the probability for axis.
        """
        v = np.zeros(self.shape)
        if isinstance(self.values[0], Distribution):
            for i, o in enumerate(self.values):
                v[i] = o.dist_max
            return v
        return self.values

    @property
    def size(self):
        """
        Get the axis size.

        Returns:
            :py:attr:`int`: Size of axis.
        """
        if isinstance(self.values[0], Distribution):
            return len(self.values)
        return self.values.size

    @property
    def shape(self):
        """
        Get the axis shape.

        Returns:
            :py:attr:`int` or :py:attr:`tuple` of :py:attr:`int`: Shape of axis.
        """
        if isinstance(self.values[0], Distribution):
            return len(self.values)
        return self.values.shape

    def pdf(self, x):
        """"
        Get the probability density function for all of the distributions in the axes.

        Args:
            x (:py:attr:`array_like`): Values to return probability of.

        Return:
            :py:attr:`array_like`: Probability.
        """
        return self.kde.pdf(x)

    def logpdf(self, x):
        """"
        Get the natural log probability density function for all of the distributions in the axes.

        Args:
            x (:py:attr:`array_like`): Values to return natural log probability of.

        Return:
            :py:attr:`array_like`: Natural log probability.
        """
        return self.kde.logpdf(x)


def _get_kde(values):
    """
    Determine the kernel density estimate for a given set of values.

    Args:
        values (:py:attr:`array_like`): Sample for kde.

    Returns:
        :py:class:`scipy.stats.kde.gaussian_kde`: Kernel density estimate for samples.
    """
    min_size = values[0].size
    for v in values:
        if v.size < min_size:
            min_size = v.size
    random_list = [np.random.choice(v.samples, size=min_size) for v in values]
    return gaussian_kde(np.vstack(random_list))



"""
The :mod:`sampling` module implements the use of a generalised MCMC (using :py:mod:`emcee`) and nested sampling (using :py:mod:`dynesty`) for the :py:class:`~uravu.relationship.Relationship` objects.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

import numpy as np
import emcee
import dynesty
from uravu import optimize
from uravu.distribution import Distribution
from dynesty import utils as dyfunc


def mcmc(relationship, prior_function=None, walkers=50, n_samples=500, n_burn=500, progress=True):
    """
    Perform MCMC to get the probability distributions for the variables of the relationship.

    Args:
        relationship (:py:class:`uravu.relationship.Relationship`): The relationship to determine the posteriors of.
        prior_function (:py:attr:`callable`, optional): The function to populated some prior distributions. Default is :func:`uravu.relationship.Relationship.prior()`.
        walkers (:py:attr:`int`, optional): Number of MCMC walkers. Default is :py:attr:`50`.
        n_samples (:py:attr:`int`, optional): Number of sample points. Default is :py:attr:`500`.
        n_burn (:py:attr:`int`, optional): Number of burn in samples. Default is :py:attr:`500`.
        progress (:py:attr:`bool`, optional): Show tqdm progress for sampling. Default is :py:attr:`True`.

    Returns:
        :py:attr:`dict`: Dictionary with the distributions as a list (:py:attr:`'distributions'`), the chain (:py:attr:`'chain'`) and the samples as an :py:attr:`array_like` (:py:attr:`'samples'`).
    """
    if prior_function is None:
        prior_function = relationship.prior

    initial_prior = np.zeros((walkers, len(relationship.variable_medians)))
    called_prior = prior_function()

    ndims = len(relationship.variable_medians)
    for i in range(ndims):
        if relationship.variable_medians[i] != 0:
            initial_prior[:, i] = relationship.variable_medians[i] + 1e-2 * np.random.randn(walkers) * relationship.variable_medians[i]
        else:
            initial_prior[:, i] = 1e-4 * np.random.randn(walkers)

    args = [relationship.function, relationship.abscissa, relationship.ordinate, called_prior]

    sampler = emcee.EnsembleSampler(walkers, ndims, ln_probability, args=args)
    sampler.run_mcmc(initial_prior, n_samples + n_burn, progress=progress)

    post_samples = sampler.get_chain(discard=n_burn).reshape((-1, ndims))

    distributions = []
    for i in range(ndims):
        distributions.append(Distribution(post_samples[:, i]))

    results = {"distributions": distributions, "chain": sampler.get_chain().reshape((-1, ndims)), "samples": post_samples}
    return results


def ln_probability(variables, function, abscissa, ordinate, priors):
    """
    Determine the natural log probability for a given set of variables, by
    summing the prior and likelihood.

    Args:
        variables (:py:attr:`array_like`): Variables for the function evaluation.
        function (:py:attr:`callable`): The function to be evaluated.
        abscissa (:py:attr:`array_like`): The abscissa values.
        ordinate (:py:attr:`array_like`): The ordinate values.
        unaccounted_uncertainty (:py:attr:`bool`): Should an unaccounted uncertainty parameter be considered.
        prior_function (:py:attr:`callable`, optional): The function to populated some prior distributions. Default is :func:`~uravu.relationship.Relationship.prior()`.

    Returns:
        :py:attr:`float`: Negative natural log-probability between model and data, considering priors.
    """
    log_prior = 0
    for i, var in enumerate(variables):
        log_prior += priors[i].logpdf(var)
    lnl = optimize.ln_likelihood(variables, function, abscissa, ordinate)
    return log_prior + lnl


def nested_sampling(relationship, prior_function=None, progress=True, dynamic=False, **kwargs):
    """
    Perform the nested sampling, or dynamic nested sampling, in order to determine the Bayesian natural log evidence. See the :py:func:`dynesty.NestedSampler.run_nested()` documentation.

    Args:
        relationship (:py:class:`~uravu.relationship.Relationship`): The relationship to estimate the evidence for.
        prior_function (:py:attr:`callable`, optional): The function to populated some prior distributions. Default is the broad uniform priors in :func:`~uravu.relationship.Relationship.prior()`.
        progress (:py:attr:`bool`, optional): Show :py:mod:`tqdm` progress for sampling. Default is :py:attr:`True`.
        dynamic (:py:attr:`bool`, optional): Should dynamic nested sampling be used?. Default is :py:attr:`False`.

    Returns:
        :py:attr:`dict`: The results from :py:func:`dynesty.NestedSampler.run_nested()`.
    """
    if prior_function is None:
        prior_function = relationship.prior
    priors = prior_function()
    nested_sampler = dynesty.NestedSampler
    if dynamic:
        nested_sampler = dynesty.DynamicNestedSampler
    logl_args = [relationship.function, relationship.abscissa, relationship.ordinate]
    sampler = nested_sampler(optimize.ln_likelihood, nested_prior, len(relationship.variables), logl_args=logl_args, ptform_args=[priors])

    sampler.run_nested(print_progress=progress, **kwargs)
    results = sampler.results
    samples = results['samples']
    weights = np.exp(results['logwt'] - results['logz'][-1])
    new_samples = dyfunc.resample_equal(samples, weights)
    distributions = []
    for i in range(new_samples.shape[1]):
        distributions.append(Distribution(new_samples[:, i]))
    results['distributions'] = distributions

    return results


def nested_prior(array, priors):
    """
    Convert to dynesty prior style from at used within :py:mod:`uravu`.

    Args:
        array (:py:attr:`array_like`): An array of random uniform numbers (0, 1]. The shape of which is M x N, where M is the number of parameters and N is the number of walkers.
        prior_function (:py:attr:`callable`, optional): The function to populated some prior distributions. Default is :func:`uravu.relationship.Relationship.prior()`.

    Returns:
        :py:attr:`array_like`: An array of random uniform numbers distributed in agreement with the priors.
    """
    broad = np.copy(array)
    for i, prior in enumerate(priors):
        broad[i] = prior.ppf(broad[i])
    return broad



"""
Figure parameters for uravu to help make nice plots.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

from collections import OrderedDict
from matplotlib import rcParams, cycler
import seaborn as sns

sns.set_palette('colorblind')
colors = sns.palettes.SEABORN_PALETTES['colorblind']

FONTSIZE = 15
NEARLY_BLACK = "#161616"
LIGHT_GREY = "#F5F5F5"
WHITE = "#ffffff"

MASTER_FORMATTING = {
    "axes.formatter.limits": (-3, 3),
    "xtick.major.pad": 7,
    "ytick.major.pad": 7,
    "ytick.color": NEARLY_BLACK,
    "xtick.color": NEARLY_BLACK,
    "axes.labelcolor": NEARLY_BLACK,
    "axes.spines.bottom": True,
    "axes.spines.left": True,
    "axes.spines.right": True,
    "axes.spines.top": True,
    "axes.axisbelow": True,
    "legend.frameon": False,
    'axes.edgecolor': NEARLY_BLACK,
    "pdf.fonttype": 42,
    "ps.fonttype": 42,
    "mathtext.fontset": "custom",
    "font.size": FONTSIZE,
    "font.family": "sans-serif",
    "font.serif" : "Helvetica",
    "savefig.bbox": "tight",
    "axes.facecolor": LIGHT_GREY,
    "axes.labelpad": 10.0,
    "axes.labelsize": FONTSIZE * 0.8,
    "axes.titlepad": 20,
    "axes.titlesize": FONTSIZE,
    "axes.grid": False,
    "grid.color": WHITE,
    "lines.markersize": 7.0,
    "lines.scale_dashes": False,
    "xtick.labelsize": FONTSIZE * 0.8,
    "ytick.labelsize": FONTSIZE * 0.8,
    "legend.fontsize": FONTSIZE * 0.8,
    "lines.linewidth": 2,
}

for k, v in MASTER_FORMATTING.items():
    rcParams[k] = v

